################################################################################
################################################################################
## TALLER: Significacin de Modelos Predictivos 
##         X Jornadas Usuarios de R. Murcia (Nov-2018)
##
## Script: Comparacin de 2 Modelos Predictivos con paralelizacin
##
## Autor: Jess Herranz. GEICAM
################################################################################
################################################################################

rm(list=ls(all=TRUE))

################################################################################
## PARAL-1: Paquete que permite la paralelizacin
library(parallel)          


################################################################################
## PARAL-2: Inicializa los Cluster
################################################################################

## Se pregunta por el nmero de cores del ordenador y se inicializan, dejando 
## un procesador libre

nc <- detectCores() 
nc      
w.cluster <- makeCluster( nc )   

## Para declarar un nmero fijo, 28, se podra poner: 
## w.cluster <- makeCluster( 28 )   


################################################################################
################################################################################
## PARAL-3: Con la funcin "clusterEvalQ" se cargan los paquetes, los datos y 
## las variables en cada Worker, que es el primer parmetro
################################################################################
################################################################################

ignore = clusterEvalQ ( w.cluster, 
{ 

################################################################################
## Libreras y Funciones
################################################################################

library(e1071)
library(pROC) 
library(caret)

## Se cargan las funciones generales
source ( "C://Taller Significacin Modelos Predictivos/Funciones Generales.r")
source ( "C://Taller Significacin Modelos Predictivos/Funciones SVM.r")  

################################################################################
## Fichero Datos: Alzheimer
################################################################################

xx = read.csv("C://Taller Significacin Modelos Predictivos/Alzheimer.csv", sep=";")
dim(xx)
#head(xx)

xx$Class = as.factor(xx$Class)

n.all = nrow(xx)                 ## Nmero de observaciones 


## Distribucin de la Variable Respuesta                       
table ( xx$Class )
prop.table ( table ( xx$Class ) )

## Nmero de la columna donde est la variable respuesta
column.Resp = which ( names(xx) == "Class" ) 

## Nmero de las columnas de los Predictores de cada uno de los modelos predictivos
names(xx)
column.Mod.BASE = 125:132         ## Modelo Base
column.Mod.ADD = 1:124            ## Variables a aadir
column.Mod.COM = c ( column.Mod.BASE , column.Mod.ADD )  ## Modelo Completo 

## Normalizacin de las Variables (con SVM es necesario)
w.trans = preProcess (  xx [ , - column.Resp ] , method = c ("center", "scale") )
xx [ , - column.Resp ]  = predict( w.trans , xx [ , - column.Resp ] )


################################################################################
## PARAL-4: Se define dentro de cada Cluster una funcin que es la que hace UNA 
## PERMUTACIN, y que ser llamado con una funcin de la familia "apply"
## - Contiene bsicamente el cdigo del "for" del bucle de permutaciones
## - El df "xx" aqu se introducir en un parmetro "par.xx"
################################################################################

SVM.Perm.Paralelo = function ( par.ind.perm , par.xx , n.cv=10 )
{

  n.all = nrow(par.xx)          ## Se ha aadido en la paralelizacin
  
  ## Permutamos las filas 
  permut = sample ( 1:n.all )                ## Permutamos las Posiciones

  ## Dataframe de trabajo. Se mantienen los predictores del modelo Base y la Respuesta
  ## Se permutar los predictores aadidos
  ww = cbind ( par.xx [        , c ( column.Mod.BASE , column.Resp ) ] , 
               par.xx [ permut , c ( column.Mod.ADD ) ] )
  names(ww)[ which ( names(ww) == "Class" ) ] = "perm.Class"    ## Cambio de nombre para clarificar


  ######################################################
  ## Proceso de CV con los predictores aadidos permutados

  ## Matriz para almacenar los errores y los AUCs
  w.error = rep ( NA, n.cv )
  w.auc   = rep ( NA, n.cv )

  ## Crea el vector "folds" que contiene el grupo (fold) de cada observacin  
  folds = sample ( rep ( 1:n.cv, length=n.all ) )
  if ( min(table (folds, ww$perm.Class )) == 0 ) folds = sample ( rep ( 1:n.cv, length=n.all ) )
  if ( min(table (folds, ww$perm.Class )) == 0 ) folds = sample ( rep ( 1:n.cv, length=n.all ) )
  
  
  ## Validacin Cruzada
  for ( ind.cv in 1:n.cv )
  {
      ## Creamos los datasets de training y testing en este fold
      cv.train = ww [ folds != ind.cv , ]
      cv.test  = ww [ folds == ind.cv , ] 

      ## Proceso con la Respuesta Permutada !!!!
      ww.mod = ProcessSVM ( cv.train, cv.test, "perm.Class" )

      ## Error Rate and AUC
      w.error [ind.cv] = ErrRate ( cv.test$perm.Class, ww.mod$pred.class )$err.rate 
      w.auc [ind.cv] = ROCAnalysis ( cv.test$perm.Class, ww.mod$pred.prob[ , "Impaired" ])$AUC
  
  } ## ind.cv


  ## Tasa de Error y AUC para la permutacin es la media de AUC y ER de la CV 
  #perm.ER[ ind.perm ]  = mean(w.error)
  #perm.AUC[ ind.perm ] = mean(w.auc)

  #######################
  ## OUTPUT: Los resultados de la permutacin son el ER y AUC
  list ( par.ind.perm = par.ind.perm, 
         out.ER = mean(w.error) , 
         out.AUC = mean(w.auc) )
} ## end-function SVM.Perm.Paralelo


## PARAL-5: aqu se cierra la funcin ""clusterEvalQ" 
  NULL      ## No devuelve nada

} ) ## final de las rdenes de R (2 parmetro) y final de la funcin "clusterEvalQ


################################################################################
## PARAL-6: EJECUCIN del Proceso
## - Aqu se ejecuta la funcin en los Clusters que estn inicializados 
################################################################################

 
#####################################################
## PARAL-7: Esta es la funcin "apply" que ejecuta la funcin en cada cluster
## repartiendo el "set.perm" entre ellos

## Conjunto de permutaciones que se van a ejecutar, en un vector
set.perm = 1:100

time1 = Sys.time()    ## para controlar el tiempo que tarda

perm.out  = clusterApply ( w.cluster, set.perm,  
              function ( w.ind ) SVM.Perm.Paralelo ( par.ind.perm = w.ind , par.xx = xx , n.cv=10 ) )

time2 = Sys.time()
difftime ( time2 , time1 , units = "mins"  )


#####################################################
## PARAL-8: MUY IMPORTANTE: Se cierran los clusters para liberar los procesadores

stopCluster(w.cluster)


#####################################################
## PARAL-9: Resultados del Proceso

## Explorando el objeto "perm.out" que devuelve: es una lista con una componente
## para cada permutacin, que contiene la salida de la funcin "SVM.Perm.Paralelo"
class( perm.out)
length ( perm.out )
perm.out[[1]]


## Para obtener vectores uniendo todos los resultados
## Estos resultados habra que salvarlos y en otro script calcular los ER y AUC
## de los datos observados
perm.ER   = as.numeric( unlist( sapply( perm.out , '[' , 'out.ER'  ))) 
perm.AUC  = as.numeric( unlist( sapply( perm.out , '[' , 'out.AUC' ))) 


## Exploramos las Tasas de Error y AUCs con las respuestas permutadss
length(perm.ER)
sort(perm.ER)[1:8]
mean(perm.ER)
length(perm.AUC)
sort(perm.AUC, decreasing=T)[1:8]
mean(perm.AUC)

dev.new()
hist( perm.AUC, main="AUC - permuted response" , xlab="AUC", breaks=20 )
dev.new()
hist( perm.ER, main="ER - permuted response" , xlab="Error Rate", breaks=20 )


